// Virvo - Virtual Reality Volume Rendering
// Copyright (C) 1999-2003 University of Stuttgart, 2004-2005 Brown University
// Contact: Jurgen P. Schulze, jschulze@ucsd.edu
//			Quei-Chun (Nancy) Hsu, ???
//			Han Suk Kim, hskim@cs.ucsd.edu
//
// This file is part of Virvo.
//
// Virvo is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library (see license.txt); if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


#ifndef __BRICK_READER_H__
#define __BRICK_READER_H__

#include "vvtoolshed.h"
#include "vvvoldesc.h"

#include <stdio.h>
#include <vector>
//#include "TIFFImage.h"
#include "xvfVolume.h"


class FileInfo
{
  public:
	FileInfo(){};
	~FileInfo(){};

  public:
	int level;
	char name[1024];
	int sizeX;
	int sizeY;
	int sizeZ;
};


class BrickReader
{
  public:
	/**
	
	  */
	BrickReader() {}

	/**
	  close files if opened
	*/
	~BrickReader();

	/** initialize file pointers 
	  call fopen(), if any of the files does not exist, raise an error
	*/
	int init(int maxMipLevel, int brickSize, std::vector<FileInfo>& _fileDescList);

	/** read data to buf
	  find an appropriate file data and read BRICK_SIZE x BRICK_SIZE x BRICK_SIZE data

	  @param level specifies mipmap level
	  @param x specifies the relative x coordinate of the brick
	  @param y specifies the relative y coordinate of the brick
	  @param z specifies the relative z coordinate of the brick
	  @param c specifies which channel to be read
	  @return buf stores the dara read from file
	  @return 0 if OK, -1 otherwise
	*/
	int getOneChannelData(int level, int x, int y, int z, int c, uchar* buf);

	/** read multi channel data to buf
	  find an appropriate file data and read c * BRICK_SIZE x BRICK_SIZE x BRICK_SIZE data

	  @param level specifies mipmap level
	  @param x specifies the relative x coordinate of the brick
	  @param y specifies the relative y coordinate of the brick
	  @param z specifies the relative z coordinate of the brick
	  @return buf stores the dara read from file
	  @return 0 if OK, -1 otherwise
	*/
	int getMultiChannelData(int level, int x, int y, int z, uchar* buf);

	vvVolDesc* getVolDesc() { return vd; };

  private:
	
	XVFVolume **mipmaps;	
	std::vector<FileInfo> fileDescList;
	vvVolDesc* vd;

};	// end of class definition


#endif


