#include "xvfVolume.h"
#include "vvtokenizer.h"
#include "vvdebugmsg.h"
#include "vvfileio.h"

XVFVolume::XVFVolume() {
	vd = new vvVolDesc();
	//vd -> setDefaults();
}

XVFVolume::XVFVolume(char* filename)
{
	vd = new vvVolDesc();

	if((fpVol = loadHeaders(filename)) == NULL) {
		vvDebugMsg::msg(1, "Error: cannot open file: ", filename);
		return;
	}
	whereData = ftell(fpVol);
}


XVFVolume::~XVFVolume() {
}

FILE * XVFVolume::loadHeaders(char *filename) {
	const char xvfID[4] = "XVF";
	FILE* fp;                                       // volume file pointer
	vvTokenizer* tok;
	vvTokenizer::TokenType ttype;                   // currently processed token type
	int i;                                          // counters
	int frameSize;                                  // size of a frame in bytes
	bool done;
	int encodedSize;                                // size of encoded data array
	vvToolshed::EndianType endian;

	vvDebugMsg::msg(1, "vvFileIO::loadXVFFile()");

	vd->setFilename(filename);
	if (vd->getFilename()==NULL) return NULL;

	if ( (fp = fopen(vd->getFilename(), "rb")) == NULL) 
	{
		vvDebugMsg::msg(1, "Error: Cannot open file.");
		return NULL;
	}

	// Read magic code:
	for (i=0; i<int(strlen(xvfID)); ++i)
	{
		if (fgetc(fp) != xvfID[i])
		{
			fclose(fp);
			// vvDebugMsg::msg(1, "Trying to load old style XVF format.");
			// return loadXVFFileOld(vd);
			vvDebugMsg::msg(1, "Old style XVF format unsupported.");
			return NULL;
		}
	}
	fseek(fp, 0, SEEK_SET);

	vd->removeSequence();                           // delete previous volume sequence

	// Read header:
	tok = new vvTokenizer(fp);
	tok->setCommentCharacter('#');
	tok->setEOLisSignificant(false);
	tok->setCaseConversion(vvTokenizer::VV_UPPER);
	tok->setParseNumbers(true);
	tok->setWhitespaceCharacter(' ');

	done = false;
	while (!done)
	{
		  // Read a token:
		ttype = tok->nextToken();
		if (ttype != vvTokenizer::VV_WORD)
		{
			cerr << "Error in line " << tok->getLineNumber() << " of XVF file" << endl;
			tok->nextLine();
		}
		else
		{
			if (strcmp(tok->sval, "XVF")==0)
			{
				tok->nextLine();
			}
			else if (strcmp(tok->sval, "VERSION")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				cerr << "Reading XVF file version " << tok->nval << endl;
			}
			else if (strcmp(tok->sval, "VOXELS")==0)
			{
				for (i=0; i<3; ++i)
				{
					ttype = tok->nextToken();
					assert(ttype == vvTokenizer::VV_NUMBER);
					vd->vox[i] = int(tok->nval);
				}
			}
			else if (strcmp(tok->sval, "TIMESTEPS")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				vd->frames = int(tok->nval);
			}
			else if (strcmp(tok->sval, "BPC")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				vd->bpc = int(tok->nval);
			}
			else if (strcmp(tok->sval, "CHANNELS")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				vd->chan = int(tok->nval);
			}
			else if (strcmp(tok->sval, "DIST")==0)
			{
				for (i=0; i<3; ++i)
				{
					ttype = tok->nextToken();
					assert(ttype == vvTokenizer::VV_NUMBER);
					vd->dist[i] = tok->nval;
				}
			}
			else if (strcmp(tok->sval, "ENDIAN")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_WORD);
				if (strcmp(tok->sval, "LITTLE")==0) endian = vvToolshed::VV_LITTLE_END;
				else endian = vvToolshed::VV_BIG_END;
			}
			else if (strcmp(tok->sval, "DTIME")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				vd->dt = tok->nval;
			}
			else if (strcmp(tok->sval, "MINMAX")==0)
			{
				for (i=0; i<2; ++i)
				{
					ttype = tok->nextToken();
					assert(ttype == vvTokenizer::VV_NUMBER);
					vd->real[i] = tok->nval;
				}
			}
			else if (strcmp(tok->sval, "POS")==0)
			{
				for (i=0; i<3; ++i)
				{
					ttype = tok->nextToken();
					assert(ttype == vvTokenizer::VV_NUMBER);
					vd->pos[i] = tok->nval;
				}
			}
			else if (strcmp(tok->sval, "CHANNELNAMES")==0)
			{
				if (vd->chan<1) tok->nextLine();
				else
				{
					for (i=0; i<vd->chan; ++i)
					{
						ttype = tok->nextToken();
						assert(ttype == vvTokenizer::VV_WORD);
						vd->setChannelName(i, tok->sval);
					}
				}
			}
			else if (strcmp(tok->sval, "TF_PYRAMID")==0)
			{
				fseek(fp, tok->getFilePos(), SEEK_SET);
				vd->tf._widgets.append(new vvTFPyramid(fp), vvSLNode<vvTFWidget*>::NORMAL_DELETE);
				tok->setFilePos(fp);
			}
			else if (strcmp(tok->sval, "TF_BELL")==0)
			{
				fseek(fp, tok->getFilePos(), SEEK_SET);
				vd->tf._widgets.append(new vvTFBell(fp), vvSLNode<vvTFWidget*>::NORMAL_DELETE);
				tok->setFilePos(fp);
			}
			else if (strcmp(tok->sval, "TF_COLOR")==0)
			{
				fseek(fp, tok->getFilePos(), SEEK_SET);
				vd->tf._widgets.append(new vvTFColor(fp), vvSLNode<vvTFWidget*>::NORMAL_DELETE);
				tok->setFilePos(fp);
			}
			else if (strcmp(tok->sval, "TF_CUSTOM")==0)
			{
				fseek(fp, tok->getFilePos(), SEEK_SET);
				vd->tf._widgets.append(new vvTFCustom(fp), vvSLNode<vvTFWidget*>::NORMAL_DELETE);
				tok->setFilePos(fp);
			}
			else if (strcmp(tok->sval, "ICON")==0)
			{
				ttype = tok->nextToken();
				assert(ttype == vvTokenizer::VV_NUMBER);
				vd->iconSize = int(tok->nval);
				tok->nextLine();

				// Read icon:
				if (vd->iconSize>0)
				{
					delete[] vd->iconData;
					int iconBytes = vd->iconSize * vd->iconSize * vvVolDesc::ICON_BPP;
					vd->iconData = new uchar[iconBytes];
					fseek(fp, tok->getFilePos(), SEEK_SET);
					int encodedSize = vvToolshed::read32(fp);
					if (encodedSize>0)                      // compressed icon?
					{
						uchar* encoded = new uchar[encodedSize];
						if ((int)fread(encoded, 1, encodedSize, fp) != encodedSize)
						{
							cerr << "Error: Insuffient compressed icon data in file." << endl;
							fclose(fp);
							delete[] vd->iconData;
							vd->iconData = NULL;
							delete[] encoded;
							return NULL;
						}
						if (vvToolshed::decodeRLE(vd->iconData, encoded, encodedSize, vvVolDesc::ICON_BPP, iconBytes) < 0)
						{
							cerr << "Error: Decoding exceeds icon size." << endl;
							fclose(fp);
							delete[] vd->iconData;
							vd->iconData = NULL;
							delete[] encoded;
							return NULL;
						}
						delete[] encoded;
					}
					else                                    // uncompressed icon
					{
						if ((int)fread(vd->iconData, 1, iconBytes, fp) != iconBytes)
						{
							cerr << "Error: Insuffient uncompressed icon data in file." << endl;
							fclose(fp);
							delete[] vd->iconData;
							vd->iconData = NULL;
							return NULL;
						}
					}
					tok->setFilePos(fp);
				}
			}
			else if (strcmp(tok->sval, "VOL_ORG")==0)
			{
				tok->nextToken();
				setVolOrganization((int)(tok->nval));
			}
			else if (strcmp(tok->sval, "VOXELDATA")==0)
			{
				tok->nextLine();
				done = true;
			}
			else
			{
				if (ttype == vvTokenizer::VV_WORD)
				{
					cerr << "Ignoring unknown header entry: " << tok->sval << endl;
				}
				tok->nextLine();
			}
		}
	}

	frameSize = vd->getFrameBytes();

	fseek(fp, tok->getFilePos(), SEEK_SET);
	encodedSize = vvToolshed::read32(fp);

	if(encodedSize>0) {
		vvDebugMsg::msg(1, "Encoded volume file unsupported!");
		fclose(fp);
		return NULL;
	}

	setBrickSize(brickSize);

	whereData = ftell(fp);
	fpVol = fp;
	return fp;
}

